;;;-*- Mode: Lisp; Package: CCL -*-

;;   Copyright (C) 2003 Sandia National Laboratories

;; Sandia National Laboratories grants you the rights to distribute
;; and use this software as governed by the terms
;; of the Lisp Lesser GNU Public License
;; (http://opensource.franz.com/preamble.html),
;; known as the LLGPL.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; Lesser General Public License for more details.


;;; search-files-browser.lisp
;;; Tuesday March 4, 2003 Shannon Spires <svspire@sandia.gov>
;;; No warranties; use at your own risk.

;;; Uses databrowser to show multi-file search results. The next time you do "Search Files",
;;;  you'll see the difference.
;;; Shows three columns:
;;;  filename plus hits and last modified date. Hits is how many times the
;;;  search string occurred in the file, with your 'search comments' choice taken into account.
;;; Also shows a 'files found' item at the bottom of the dialog that shows how many files
;;;  contained the string.

;;; You can re-sort the items by clicking the column headers, change the width of the columns
;;;   by dragging them, and even drag entire columns around. Note what happens if you
;;;   widen the 'Last Modifed' column--it shows more information.

;;; BUGS:
;;; Hits column sometimes gets "..." instead of actual hits. This usually happens when its
;;; column starts out too small. Don't know how to fix it.

(in-package :ccl)
(require :databrowser)

; Following duplicates static-text-control-item.lisp patch
(require :appearance-manager "ccl:examples;appearance-manager-folder;appearance-manager.lisp")
(require :new-control-dialog-item)

(export 'static-text-control-item)

(defclass static-text-control-item (new-control-dialog-item) ; SVS modified for Terje's appearance-manager
  ((procid :initarg :procid :initform #$kControlStaticTextProc
           :allocation :class)))

; Draws gray over theme-background windows. Should be transparent. Don't know how to fix it.
(defmethod install-view-in-window :after ((item static-text-control-item) dialog)
  (declare (ignore dialog))
  (with-cstrs ((str (dialog-item-text item)))
    (#_SetControlData (dialog-item-handle item) #$kControlEntireControl
     #$kControlStaticTextTextTag (length (dialog-item-text item)) str)
    ;(#_SetUpControlBackground (dialog-item-handle item) 32 t) ; doesn't seem to do anything
    ))

(defmethod view-draw-contents :before ((item static-text-control-item))
  (with-cstrs ((str (dialog-item-text item)))
    (#_SetControlData (dialog-item-handle item) #$kControlEntireControl
     #$kControlStaticTextTextTag (length (dialog-item-text item)) str)))

; end of static-text-control-item.lisp patch

(defclass brw::file-collection-databrowser (brw:collection-databrowser)
  ())

(defclass db-item-selection-dialog-with-stop (item-selection-dialog-with-stop)
  ())


#|
(defmethod window-null-event-handler :after ((me db-item-selection-dialog-with-stop))
  (let* ((fc (view-named 'file-count me))
         (current-count (brw::get-row-count (view-named 'sequence-dialog-item me)))
         (current-displayed-count (ignore-errors (parse-integer (dialog-item-text fc)))))
    (unless (eql current-displayed-count current-count) ; prevent flickering
      (set-dialog-item-text fc
                            (format nil "~D" current-count)))))

|#

#|
(defmethod set-view-size :after ((window item-selection-dialog-with-stop) h &optional v)
  (declare (ignore h v))
  ;; prevent a re-size while resizing - the minimum size code depends on the current size
  (let ((fc-prompt (view-named 'files-found window))
        (fc (view-named 'file-count window)))
    (set-view-position fc-prompt (point-h (view-position fc-prompt)) (- (point-v (view-size window)) 30))
    (set-view-position fc (point-h (view-position fc)) (- (point-v (view-size window)) 30))))
|#

(defmethod add-dialog-items ((dialog db-item-selection-dialog-with-stop)
                             &rest args &key the-list selection-type action-function
                             default-button-text stop-button-text column-descriptors)
  (declare (ignore args the-list))
  (flet ((act-on-items (item)
           (declare (ignore item))
           (let ((db-item (view-named 'sequence-dialog-item dialog)))
             (funcall action-function
                      (mapcar #'(lambda (rowid) (brw:lookup-by-row db-item rowid))
                              (brw:selected-rows db-item))))))
    
    (let ((button-width (if (osx-p) 60 58))
          (button-height 20)
          (state-width 60)
          (text-height 12)
          (spacing-a 13)
          (border 3)
          (size (view-size dialog)))
      (add-subviews dialog
                    (make-instance 'static-text-control-item-ar
                      :dialog-item-text ""
                      :view-position (make-point (+ spacing-a border)
                                                 (+ (floor (- button-height text-height) 2)
                                                    (- spacing-a border)))
                      :view-size (make-point state-width text-height)
                      :view-font '("Geneva" 9 :bold)
                      :view-nick-name 'state)
                    (make-instance 'ellipsized-text-dialog-item 
                      :dialog-item-text ""
                      :view-position (make-point (+ spacing-a border border state-width)
                                                 (+ (floor (- button-height text-height) 2)
                                                    (- spacing-a border)))
                      :view-size (make-point (- (point-h size) 
                                                (+ (+ spacing-a border) border state-width)
                                                border
                                                button-width (- spacing-a border) spacing-a)     ; button left
                                             text-height)
                      :view-font '("Geneva" 9 :italic)
                      :view-nick-name 'detailed-state
                      :compress-text nil)       ; avoid bug where compressing switches to non-italics
                    (make-instance 'button-dialog-item
                      :dialog-item-text stop-button-text
                      :view-position (make-point (- (point-h size) (- spacing-a border) spacing-a button-width)
                                                 (- spacing-a border))
                      :view-size (make-point button-width button-height)
                      :view-nick-name 'stop-button
                      :dialog-item-action
                      #'(lambda (item)
                          (declare (ignore item))
                          (when (stop-action dialog)
                            (funcall (stop-action dialog) dialog))))
                    (make-instance
                      'brw::file-collection-databrowser
                      :view-position (make-point spacing-a 
                                                 (+ (- spacing-a border) button-height spacing-a))
                      :view-size (make-point (- (point-h size)
                                                spacing-a     ; left
                                                (- spacing-a border) spacing-a)          ; right
                                             (- (point-v size)
                                                (- spacing-a border) button-height spacing-a     ; top
                                                spacing-a button-height spacing-a))     ; bottom
                      :VIEW-FONT ;'("Monaco" 9 :extend) ; seems to be a bug here--need :extend or it condenses the text
                      '(11)
                      :selection-type selection-type
                      :view-nick-name 'sequence-dialog-item ; keep the same old name as before
                      :vscrollp t
                      :hscrollp nil
                      :column-descriptors column-descriptors
                      )                    
                    (MAKE-DIALOG-ITEM
                     'STATIC-TEXT-CONTROL-ITEM-AR
                     (make-point 20 (- (point-v (view-size dialog)) 30))
                     #@(98 16)
                     "Files found:"
                     'NIL
                     :VIEW-NICK-NAME
                     'FILES-FOUND
                     :VIEW-FONT
                     '("Geneva" 9 :SRCOR :BOLD (:COLOR-INDEX 0)))
                    (MAKE-DIALOG-ITEM
                     'STATIC-TEXT-CONTROL-ITEM-AR
                     (make-point 82 (- (point-v (view-size dialog)) 30))
                     #@(54 16)
                     "0"
                     'NIL
                     :VIEW-NICK-NAME
                     'file-COUNT
                     :VIEW-FONT
                     '("Geneva" 9 :SRCOR :PLAIN (:COLOR-INDEX 0)))
                    (make-instance 
                      'default-button-dialog-item
                      :dialog-item-enabled-p nil ; until something is selected
                      :dialog-item-text default-button-text
                      :view-position (make-point (- (point-h size) (- spacing-a border) spacing-a button-width)
                                                 (- (point-v size) spacing-a button-height))
                      :view-size (make-point button-width button-height)
                      :view-nick-name 'locate-button
                      :dialog-item-action
                      #'act-on-items)))
    nil))

(defun db-select-items-from-list-with-stop (the-list &key (window-title "Select an Item")
                                                  (selection-type :single)
                                                  (action-function #'identity)
                                                  (stop-button-text "Stop")
                                                  (stop-action nil)
                                                  (default-button-text "Find It")
                                                  (window-position-variable nil)
                                                  (window-size-variable nil)
                                                  (column-descriptors (list 'identity))
                                                  )
  "Displays the elements of a list, and returns the item chosen by the user"
  #+ccl-3
  (when stop-action
    (let ((process *current-process*)
          (user-stop-action stop-action))
      (setq stop-action
            #'(lambda (dialog)
                (process-interrupt process user-stop-action dialog)))))
  (let ((dialog (make-instance
                  'db-item-selection-dialog-with-stop
                  :stop-action stop-action
                 ; :back-color *tool-back-color*
                 ; :content-color *tool-back-color*
                  :theme-background t
                  :window-show t
                  :window-position-variable window-position-variable
                  :window-size-variable window-size-variable
                  :window-title window-title
                  :view-size (or (and window-size-variable (symbol-value window-size-variable))
                                 #@(400 180))
                  :view-position  (or (and window-position-variable (symbol-value window-position-variable))
                                      (make-point (%ilsr 1 (- *screen-width* 400))
                                                  90))
                  )))
    (add-dialog-items dialog
                      :the-list the-list
                      :selection-type selection-type
                      :action-function action-function
                      :stop-button-text stop-button-text
                      :default-button-text default-button-text
                      :column-descriptors column-descriptors)
    (brw::auto-size-columns (view-named 'sequence-dialog-item dialog))
    (window-show dialog)))

(defclass file-result-object ()
  ((pathname :initarg :pathname :initform nil :accessor get-pathname)
   (mod-date :initarg :mod-date :initform nil :accessor mod-date)
   (hits :initarg :hits :initform 0 :accessor hits
         :documentation "How may occurrences of the given string were in the file.")))

(defun db-searching-files-selection-dialog (string)
  (db-select-items-from-list-with-stop nil
                                       :window-title (format nil "Files Containing: ~A" string)
                                       :window-position-variable '*search-files-dialog-position*
                                       :window-size-variable '*search-files-dialog-size*
                                       :stop-action
                                       #'(lambda (dialog)
                                           (dialog-item-disable (view-named 'stop-button dialog))
                                           (set-dialog-item-text (view-named 'detailed-state dialog) "Aborted.")
                                           (abort))
                                       :action-function
                                       #'(lambda (list)
                                           (using-binder-editor
                                             (maybe-start-isearch (ed (get-pathname (car list))) string)))
                                       :column-descriptors (vector
                                                            (list #'(lambda (item comparing-p browser rowID)
                                                                  (declare (ignore browser rowID))
                                                                  (if comparing-p ; compare based only on filename portion
                                                                    (file-namestring (get-pathname item))
                                                                    (get-pathname item)))
                                                              :title "Filename" :maxwidth 800 :justification :left :more-function-parameters t)
                                                             '(hits :title "Hits" ) ; :minwidth 50)
                                                             '(mod-date :title "Last Modified" :property-type :time :minwidth 100)
                                                             )
                                       ))

(defun db-searching-files-dialog (string files-function pathname search-comments-p)
  (let* ((dialog (db-searching-files-selection-dialog string))
         (db (view-named 'sequence-dialog-item dialog))
         (state (view-named 'state dialog))
         (detailed-state (view-named 'detailed-state dialog))
         (complete? nil)
         (files nil)
         (current-file nil)
         (filecount 0)
         (initialized-buffer nil)
         (scratch-buffer nil))
    (set-dialog-item-text state "Searching:")
    (set-dialog-item-text detailed-state "Collecting File List")
    (setq files (funcall files-function))
    (unless files 
      (window-close dialog)
      (return-from db-searching-files-dialog))

    (unwind-protect
      (flet ((start-file-search (file)
               (when (null (wptr dialog))
                 (return-from db-searching-files-dialog))
               (set-dialog-item-text detailed-state (format nil "~A" file))
               )
             (search-successful (file)
               (cond (current-file
                      (incf (hits current-file)))
                     (t ; no current-file yet
                      (setf current-file (make-instance 'file-result-object
                                           :pathname file
                                           :mod-date (file-write-date file)
                                           :hits 1))
                      (brw:add-items db (list current-file))
                      (set-dialog-item-text (view-named 'file-count dialog) (format nil "~D" (incf filecount)))
                      ))
               )
             (ensure-buffer (file)
               (or initialized-buffer
                   (progn
                     (setf initialized-buffer (or scratch-buffer (setq scratch-buffer (make-buffer))))
                     (%buffer-insert-file initialized-buffer file 0)
                     initialized-buffer)))
             )
        
        (declare (dynamic-extent #'start-file-search #'search-successful #'ensure-buffer))
        (flet ((f (file pos)
                 (when file
                   (cond ((eq pos t) (start-file-search file))
                         (pos ;Found, but may be in a comment.
                          (when (or search-comments-p
                                    (not (buffer-point-in-comment? (ensure-buffer file) pos 0 (buffer-size (ensure-buffer file))))
                                    )
                            (search-successful file))
                          t ; keep searching in this file
                          )
                         (t ; pos was nil so we're done with this file
                          (when current-file
                            (brw:update db (brw:lookup-by-object db current-file)) ; so hits will be updated
                            (setf current-file nil))
                          (when scratch-buffer (buffer-delete scratch-buffer 0 t))
                          (setf initialized-buffer nil)
                          )
                         ))))
          (declare (dynamic-extent #'f))
          (bm-find-string-in-files string files #'f))
        
        (setq complete? t))
      (ignore-errors
       (brw:update db t) ; update the whole thing
       (dialog-item-disable (view-named 'stop-button dialog))
       (set-dialog-item-text detailed-state (if complete? 
                                              (format nil "Completed searching ~A." pathname)
                                              "Canceled."))))
    ))

; Handle double clicks
(defmethod brw:databrowser-item-double-clicked ((browser brw::file-collection-databrowser) rowID)
  (declare (ignore rowID))
  (let ((default-button (default-button (view-container browser))))
    (when (dialog-item-enabled-p default-button)
      (dialog-item-action default-button))))

(defmethod brw:databrowser-item-selected ((browser brw::file-collection-databrowser) rowID)
  (declare (ignore rowID))
  (dialog-item-enable (default-button (view-container browser))))

; Doesn't seem to be any easy way to deselect all rows, so we don't worry about disabling default button

; Handle <return> key
(defmethod view-key-event-handler ((browser brw::file-collection-databrowser) char)
  (case char
    ((#\newline)
     (let ((selected-items (brw:selected-rows browser)))
       (brw:databrowser-item-double-clicked browser (car selected-items))
       ))
    (t
     (call-next-method) ; handle arrow keys etc.
     )))

(defvar *old-searching-files-dialog* (fdefinition 'searching-files-dialog))

(eval-when (:load-toplevel :execute)
(let ((*warn-if-redefine-kernel*))
  (setf (fdefinition 'searching-files-dialog) #'(lambda (&rest args) (apply 'db-searching-files-dialog args)))))

; to restore the old dialog
#|
(let ((*warn-if-redefine-kernel*))
  (setf (fdefinition 'searching-files-dialog) *old-searching-files-dialog*))
|#