;-*- Mode: Lisp; Package: CCL -*-

;;;;;;;;;;;;;;;;;;;;;
;;
;; make-ccl-ballon-help.lisp
;;
;; Create balloon help resources for MCL 2.0
;;

;;;;;;;;;;;;;;;;;;;;;
;;
;; Mod History
;;
;; 04/17/97 bill "Create Bug Report" menu item
;; ------------- 4.1b2
;; 04/08/97 bill Rename the "Eval" menu to "Lisp".
;; ------------- 4.1b1
;; 10/01/96 slh  comment out Environment dialog strings; new docs -> fred format files
;;  9/23/96 slh  new docs: *autoclose-inactive-listeners*, *always-eval-user-defvars*
;; 04/25/96 bill .fasl => .pfsl for ppc target
;; ------------- MCL-PPC 3.9
;; 02/24/92 bill close-help-resource-file after modifying it.
;; 11/06/91 bill Update with Sarah's help strings.
;; 07/09/91 bill New File
;;

(in-package :ccl)

(eval-when (compile load eval)
  (require :help-manager))

(defvar *balloon-help-specs*)

(defun make-ccl-balloon-help (&optional (specs *balloon-help-specs*))
  (declare (special *balloon-help-specs*))
  (unless (help-resource-file-refnum *application*)
    (let ((refnum (open-resource-file (help-resource-file *application*)
                                      :if-does-not-exist :create)))
      (when refnum (close-resource-file refnum))))
  (unless (help-resource-file-refnum *application*)
    (error "Can't create help resource file."))
  (dolist (spec specs)
    (add-balloon-help spec))
  (close-help-resource-file))

(defun add-balloon-help-resource (resource type id)
  (use-resource-file (help-resource-file-refnum *application*))
  (delete-resource type id t)
  (add-resource resource type id)  ; :attributes #$resPurgeable)
  (write-resource resource)
  (detach-resource resource)
  (#_DisposHandle resource))

(defun add-balloon-help (spec)
  (if (atom (car spec))            ; a 'STR ' resource
    (destructuring-bind (id string) spec
      (let ((handle (#_NewHandle (1+ (length string)))))
        (%put-string handle string)
        (add-balloon-help-resource handle :|STR | id)))
    ; a 'STR#' resource
    (let* ((id (caar spec))
           (string-count (length (cdr spec)))
           (size (+ 2 string-count (apply #'+ (mapcar #'length (cdr spec)))))
           (handle (#_NewHandle size))
           (offset 2))
      (%hput-word handle string-count)
      (dolist (string (cdr spec))
        (%put-string handle string offset)
        (setq offset (+ offset (length string) 1)))
      (add-balloon-help-resource handle :|STR#| id))))

(defparameter *balloon-help-specs*
  ; "About Macintosh Common Lisp" on the Apple menu.
  '(((1000)
     "Pop up the about box for Macintosh Common Lisp."
     "Pop up the about box for Macintosh Common Lisp.
The diamond next to \"About \" means that MCL is not idle.")

    ; The "File" menu
    (1100 "File menu

Use this menu to open, close & print windows and to quit the application.")
    ; "New" on the "File" menu
    (1101 "Creates a new editor window, not initially associated with any file.")
    ; "Open" on the "File" menu
    (1102 "Allows you to select a text file and creates a new editor window for that file.")
    ; "Open Selected File" on the "File" menu
    ((1103)
     "Tries to parse the selection in the top editor buffer as a pathname.If successful, it creates an editor window for the pathnames file."
     "Tries to parse the selection in the top editor buffer as a pathname.If successful, it creates an editor window for the pathnames file.
Unavailable because there is no selection in the top editor buffer.")
    ; "Close" on the "File" menu
    (1104 "Closes the current window. If the contents of the window have been changed since they were saved, a dialog box asks you if you want to save or throw away the changes.")
    ; "Save" on the "File" menu
    ((1105)
     "Saves the contents of the active window into the file named in the title bar. If the window is not associated with a disk file, Save acts like Save As."
     "Saves the contents of the active window into the file named in the title bar. If the window is not associated with a disk file, Save acts like Save As.
Unavailable because the window does not need or does not support saving.")
    ; "Save As" on the "File" menu.
    ((1106)
     "Saves the contents of the active window to a directory and filename you specify. If the file already exists, asks you to confirm if you want to write over its contents. Subsequent saves of the window contents save to the new file."
     "Saves the contents of the active window to a directory and filename you specify.
Unavailable because the window does not need or does not support saving.")
    ; "Save Copy As" on the "File" menu
    ((1107)
     "Copies the contents of the active window to a directory and filename you specify. If the file already exists, asks you to confirm if you want to write over its contents."
     "Copies the contents of the active window to a directory and filename you specify. If the file already exists, asks you to confirm if you want to write over its contents.
Unavailable because the window does not support saving")
    ; "Revert" on the "File" menu
    ((1108)
     "Reverts the contents of the window to the contents of the file on disk. All changes since the file was last saved are lost. Asks for confirmation before it reverts."
     "Reverts the contents of the window to the contents of the file on disk. All changes since the file was last saved are lost. Asks for confirmation before it reverts.
Unavailable because the window does not need or does not support reverting.")
    ; "Page Setup" on the "File" menu
    (1109 "Allows you to set printing options for the current printer.")
    ; "Print" on the "File" menu
    ((1110)
     "Prints the contents of the active window to the currently selected printer."
     "Prints the contents of the active window to the currently selected printer.
Unavailable because the window does not support printing.")
    ; "Quit" on the "File" menu
    (1111 "Closes all windows and exits from MCL. If the contents of any window have not been saved to disk, asks if you want to save them.")

    ; The window-ask-save dialog box
    (11061 "This dialog box asks you whether you want to save the contents of the window you are closing.")
    (11062 "Click this button to save the contents of the window you are closing.")
    (11063 "Click this button to close the window without saving its contents. Any changes you have made since the last save will be lost.")
    (11064 "Click this button to cancel the operation. You neither close the window nor save its contents.")

    ; The "Edit" menu
    (1200 "Edit menu

Use this menu to perform editing operations on windows and to undo your last action.")
    ; "Undo" on the "Edit" menu
    ((1201)
     "Undoes the last editor command that inserted or deleted text in the topmost window.  Successive single character deletions or insertions are considered to be a single command. Commands that do not change the window contents cannot be undone."
     "Undoes the last editor command that inserted or deleted text in the topmost window.
Not available because your last editor command cannot be undone.")
    ; "Undo more" on the "Edit" menu
    ((1202)
     "Undoes the next most recent editor command. Active immediately after an Undo has been performed."
     "Undoes the next most recent editor command. Active immediately after an Undo has been performed. Not available because your last editor command was not an Undo."
     "Redoes an Undone editor command."
     "Redoes an Undone editor command. Not available because your last editor command was not an Undo or Undo More.")

    ; "Cut", "Copy", "Paste", "Delete", "Select All" on the "Edit" menu
    ((1203)
     "Deletes the selected text and places it in the Clipboard and onto the kill ring."
     "Deletes the selected text and places it in the Clipboard and onto the kill ring.
Not available because no text has been selected or the active window does not support cutting.")
    ((1204)
     "Copies the selected text to the Clipboard and to the kill ring."
     "Copies the selected text to the Clipboard and to the kill ring.
Not available because no text has been selected or the active window does not support copying.")
    ((1205)
     "Replaces the selected text with the text from the Clipboard. If there is no current selection, inserts the text at the cursor location."
     "Replaces the selected text with the text from the Clipboard. If there is no current selection, inserts the text at the cursor location.
Not available because there is nothing in the Clipboard or the active window does not support pasting.")
    ((1206)
     "Deletes the current selection. The deleted text is not copied to the Clipboard. However, like all deleted text, it is pushed onto the kill ring."
     "Deletes the current selection. The deleted text is not copied to the Clipboard. However, like all deleted text, it is pushed onto the kill ring.
Not available because no text has been selected or the active window does not support deletion.")
    ((1207)
     "Selects the entire contents of the active window."
     "Selects the entire contents of the active window.
Not available because the window does not support this operation.")
    ; "Search" on the "Edit" menu
    (1208 "Brings up the String Search dialog box, which prompts you for a string and searches for the string in the active window. You can also choose to replace the string with another. Freds Control-S and Control-R commands can also be used for searching.")

    ; The "String Search" dialog.
    (12081 "This dialog box allows you to search for and replace strings.")
    (12082 "Type the string to search for or the string to be replaced.")
    (12083 "Leave this field blank if you simply want to search for strings.If you want to replace one string with another, type the new string here.")
    (12084 "Click this button to start searching forward.")
    (12085 "Click this button to start searching in reverse.")
    (12086 "Click this button to search from the top of the editor buffer.")
    (12087 "Click this button to replace this occurrence of the search string with the replacement string.")
    (12088 "Click this button to replace this occurrence of the search string with the replacement string and to look for the next occurrence of the search string.")
    (12089 "Click this button to replace all occurrences of the search string with the replacement string.")

    ; The "Lisp Menu"
    (1300 "Lisp menu

Use this menu to evaluate buffers, compile or load files, or interact with the break loop.")
    ; "Execute Selection" on the "Lisp" menu
    ((1301)
     "Executes the current selection in the top editor window. If there is no selection, the expression adjacent to the cursor is evaluated."
     "Executes the current selection in the top editor window. If there is no selection, the expression adjacent to the cursor is evaluated.
Not available because the window does not support evaluation.")
    ; "Execute All" on the "Eval" menu
    ((1302)
     "Executes the entire contents of the top editor window."
     "Executes the entire contents of the top editor window.
Not available because the window does not support evaluation.")
    ; "Load" on the "Eval" menu
     (1303 #-ppc-target "Prompts you to select a file. MCL then evaluates each expression in that file. You may load both .lisp and .fasl (uncompiled and compiled) files."
           #+ppc-target "Prompts you to select a file. MCL then evaluates each expression in that file. You may load both .lisp and .pfsl (uncompiled and compiled) files.")
     ; "Compile File" on the "Eval" menu
     (1304 "Prompts you for a source file and the name of a destination file, then compiles the source file into the destination file.")
     ; "Abort" on the "Eval" menu
     (1305 "Cancels the current computation and returns to the read-eval-print loop. If MCL was in a break loop, it leaves the loop. The computation cannot be resumed.")
     ; "Break" on the "Eval" menu
     (1306 "Suspends the current computation and enters a break loop. You can resume the computation by choosing Continue or by calling the function CONTINUE.")
     ; "Continue" on the "Eval" menu
     ((1307)
      "Continues the last operation halted by a break or a continuable error."
      "Continues the last operation halted by a break or a continuable error.
Not available because no operation can be continued.")
     ; "Restarts" on the "Eval" menu
     ((1308)
      "Provides a list of possible ways to restart the current operation."
      "Provides a list of possible ways to restart the current operation.
Not available because no operation can currently be restarted.")


     ; "Tools" menu
     (1400 "Tools menu

The items on this menu are programming tools. They include tools to search for symbols, edit definitions, inspect MCL objects, search sets of files, do a stack backtrace, list editor commands, and set printing or environment options.")
     ; "Apropos"
     (1401 "Prompts you for a string and displays all symbols whose names contain that string. When you select a symbol by double-clicking it, an Inspector window opens displaying information about that symbol.")
     ; "Documentation"
     (1402 "Prompts for a symbol and displays its online documentation , if available. The corresponding Fred command is Control-X Control-D.")
     ; "Edit Definition"
     (1403 "Prompts you for a symbol. Finds the file containing the definition of the symbol, opens the file in an editor window, and displays the definition. If there is more than one definition, a dialog allows selection of one of them. Fred command is Meta-period.")
     ; "Inspect"
     (1404 "Displays a window that includes Inspector Help, the Inspector History, and a list of kinds of system data you can inspect. Clicking any item opens an Inspector window.")
     ; "List Definitions"
     ((1405)
      "Lists the symbols defined at top level in the front window, allowing you to scan and search them."
      "Lists the symbols defined at top level in the front window, allowing you to scan and search them.
Not available because the window does not support this operation.")
     ; "Search Files"
     (1406 "Brings up a dialog box allowing you to search a selected group of files for a string.")
     ; "Backtrace"
     ((1407)
      "Lets you examine the state of the suspended program from within a break loop."
      "Lets you examine the state of the suspended program from within a break loop.
Not available because the program is currently not within a break loop.")
     ; "Fred Commands"
     (1408 "Lists all editor commands and their keyboard equivalents. You can search for commands with Control-S.")
     ; "Listener Commands"
     (1411 "Lists all Listener commands that differ from Fred commands, and their keyboard equivalents. You can search for commands with Control-S.")
     ; "Print Options"
     (1409 "Brings up a dialog box allowing you to change global variables that affect the Common Lisp PRINT function.")
     ; "Environment"
     (1410 "Brings up a dialog box allowing you to change global variables that affect the Macintosh Common Lisp programming environment.")
     ; "Create Bug Report"
     ((1411)
      "Creates a bug report for one of the suspended processes."
      "Creates a bug report for one of the suspended processes.
Not available because no process is in a break loop.")
      
     (1500 "Windows menu

Use this menu to choose a window to bring to the front.")
     ((1501)
      "Bring this window to the front."
      "Not available because this window is already in the front."
      "Bring this window to the front.
The cross to the left of the window name means that the window needs to be saved.")

     ; The "Apropos" dialog
     (14010 "Type a string in the box and click the \"Apropos\" button to list all symbols that contain then string.")
     ; The editable-text-dialog-item
     (14011 "Type a string here and click the \"Apropos\" button to list all symbols that contain the string.")
     ; "Apropos" button
     (14012 "List below all the symbols that contain the string in the box above.  Since this is a default button, you can press the \"Return\" or \"Enter\" key instead of clicking here.")
     ; "Inspect" button
     (14013 "Inspect the symbol that is selected in the table below.  Double clicking on a symbol will also inspect it.")
     ; The table of symbols
     (14014 "Double click on a symbol to inspect it.")     

     ; The "List Definitions" dialog
     (14060 "This window lists the symbols defined at top level in an editor window.  Double click on one or select it and click the \"Go To Def\" button to find its source code.")
     ; "Buffer Oders" radio button
     (14061 "Lists the definitions in the order in which they appear in the editor buffer.")
     ; "Alphabetical Order" radio button
     (14062 "Lists the definitions in alphabetical order.")
     ; "Go To Def" button
     (14063 "Brings the editor window to the front and positions the cursor at the beginning of the selected definition.")
     ; "Rescan" button
     (14064 "Rescans the editor buffer for its definitions. Useful if you have changed something since the last rescan.")
     ; "Filter" button
     (14065 "Click here to bring up a dialog asking for a filter string. Only symbols that contain the string will be listed.")

     ; The backtrace window
     (14070 "This is a backtrace window. It can be used to examine the state of MCL's execution stack.")
     ; Top pane
     (14071 "This pane lists the names of the functions that are pending on the stack.  Click on one to display the variables in its stack frame in the lower pane.  Double click to inspect the function itself.")
     (14072 "This pane lists the values in the stack frame for the function that is highlighted in the top pane.  To change a value, click on it and choose the Edit command from the Commands pop-up menu.  To inspect a value, double click on it.")
     ; "Frame Size" display
     (14073 "The number of values that are in the stack frame displayed in the lower pane.  This count does not include values that are stored in registers.")
     ; "Frame Address" display
     (14074 "The memory address of the control stack entry for the function highlighted in the top pane.")
     ; "Program Counter" display
     (14075 "The program counter value at which execution will continue when control returns to the function highlighted in the top pane")
     ; "Commands" menu
     (14076 "This pop up menu contains commands to edit a value in the lower pane, return from or restart the selected frame, or display a list of restarts.")
     ; "Edit" menu item
     ((14077)
      "Pop up a dialog box allowing modification of the value selected in the lower pane."
      "Pop up a dialog box allowing modification of the value selected in the lower pane. Not available because no value is selected in the lower pane.")
     ; "Return from frame" menu item
     ((14078)
      "Pop up a dialog box allowing specification of a value to return from this frame."
      "Pop up a dialog box allowing specification of a value to return from this stack frame. Not available because there is not enough information to restore the saved register context.")
     ; "Restart frame" menu item
     ((14079)
      "Pop up a dialog box to restart this frame after optionally changing its parameters."
      "Pop up a dialog box to restart this frame after optionally changing its parameters. Not available because there is not enough information to restore the saved register context.")
     (14080 "Pop up a menu of pending restarts.  Same as the \"Restarts\" command on the Eval menu.")

     ; The "Inspector Central" window
     ; "Help"
     (14041 "Opens the Inspector Help file, which contains information about shortcuts for inspector windows.")
     ; "Inspector History"
     (14042 "Opens the Inspector History window showing a history of the last few objects that were inspected.  The length of the history is initially 0.  The Inspector History window will tell you how to set the length of the history.")
     ; "Disk Devices"
     (14043 "Displays an Inspector window listing the names of all currently active disk devices.")
     (14044 "Displays an Inspector window listing all logical directory names and their physical equivalents.")
     (14045 "Displays an Inspector window listing all logical directory names and their physical equivalents.")
     (14046 "Displays an Inspector window listing all currently open MCL windows.")
     (14047 "Inspects the front window")
     (14048 "Displays an Inspector  window listing the value of the Common Lisp variable *features*, that is, listing all features provided by the implementation. This list is used by the #+ and #- reader macros.")
     (14049 "Displays an Inspector window listing all packages in MCL.")
     (14050 "Inspects the current package: the value of the symbol *package*")
     (14051 "Displays an Inspector window listing the macro characters in the current readtable. By double-clicking one you can inspect the macro dispatch characters.")
     (14052 "Displays an inspector window describing all the currently loaded record types.")
     (14053 "Displays an inspector window describing all the defined record field types")

     ; The "Print Options" dialog
     (14090 "The items on this window can be used to change the operation of the Common Lisp PRINT function.")
     ; *print-pretty*
     (14091 "When this variable is checked, the printer inserts extra whitespace to make Lisp expressions more readable.")
     ; *print-array*
     (14092 "When this variable is checked, the contents of non-string arrays are printed using #(, #*, or #nA syntax. When it is not, arrays are printed in a concise form (using #<) that is not readable.")
     ; *print-circle*
     (14093 "When this variable is checked, the printer detects circularity and shared sub structure in the structure to be printed. It uses #n= and #n# syntax to indicate the circularities.")
     ; *print-structure*
     (14094 "When this variable is checked, the printer prints structures readably, using #S syntax.")
     ; *print-escape*
     (14095 "When this variable is checked, an attempt is made to print an expression in such a way that it can be read again to produce an equal structure.")
     ; *print-gensym*
     (14096 "When this variable is checked, the printer prints the symbol #: before symbols that have no home package.")
     ; *print-radix*
     (14097 "When this variable is checked, the printer prints a radix specifier to indicate the radix in which it is printing a rational number.")
     ; *print-readably*
     (14098 "When this variable is checked, an error will be signalled if the printer attempts to print an object whose print representation is not readable by the reader.")
     ; *print-length*
     (14099 "This variable controls how many elements at a given level are printed. A value of nil, the default value, indicates that there is no limit to the number of components printed.")
     ; *print-level*
     (14100 "This variable controls how many levels deep a nested object will print. The default value is nil, which means that no control is exercised.")
     ; *print-base*
     (14101 "The value of this variable determines the radix in which the printer will print rational numbers. The value may be any integer from 2 to 36; the default value is 10.")
     ; *print-right-margin*
     (14102 "The value of this variable specifies the size of the right margin. When it is nil (the default), the right margin is set at the maximum line length that the output stream can display without wraparound or truncation.")
     ; *print-case*
     (14103 "This variable controls the case in which to print uppercase characters in the names of symbols when vertical-bar syntax is not used. Its value can be :upcase, :downcase, :capitalize, or :studly.")
     ; :upcase radio button
     (14104 "Setting *print-case* to :upcase causes unescaped symbols to be printed in UPPER CASE.")
     ; :downcase radio button
     (14105 "Setting *print-case* to :downcase causes unescaped symbols to be printed in lower case.")
     ; :capitalize radio button
     (14106 "Setting *print-case* to :downcase causes unescaped symbols To Be Capitalized")
     (14107 "Clicking this button will change the global variables as specified.")
     (14108 "Clicking this button will cancel any changes you've made to this dialog.  You expected maybe something more exciting?")

     #| obsolete - doc strings go in "ccl:help-maker;fred format;*" files
     ; The "Environment" dialog
     (14200 "The items in this window can be used to change global variables affecting various parts of the MCL environment")
     ; *emacs-mode*
     (14201 "Useful for keyboards with no Control key. When this variable is checked, the Command key means Control and Command-Shift means Command. When it is not checked, Command means Command; Command-Shift and the Control key mean Control.")
     ; *record-source-file*
     (14202 "When this variable is checked, the compiler records the source file of all definitions. You can retrieve the source by typing Meta-period when the insertion point is in a symbol, with the Inspector, or by choosing Edit Definition from the Tools menu.")
     ; *arglist-on-space*
     (14203 "When this variable is checked and you type an open parenthesis, the symbol naming a definition, and a space, the argument list of the definition is printed in the mini-buffer.")
     ; *autoload-traps*
     (14204 "When this variable is checked, the file containing the definition of a trap is automatically loaded when the trap is referenced by the #_ or #$ reader macros.")
     ; *save-doc-strings*
     (14205 "When this variable is checked, documentation strings of definitions are retained; otherwise they are discarded. Unchecking this variable can save memory.")
     ; *save-definitions*
     (14206 "When this variable is checked, the lambda expression used to generate a function is retained. This allows compiled functions to be uncompiled so that they can be stepped. When it is unchecked, the lambda expression is not saved (which saves space).")
     ; *compile-definitions*
     (14207 "When this variable is checked, functions are compiled. If it is not checked, they are evaluated.")
     ; *save-local-symbols*
     (14208 "When this variable is checked, names of arguments and local variables are retained when functions are compiled interactively. This allows the debugger to display the names of arguments and variables in the Stack Backtrace, but uses more memory.")
     ; *paste-with-styles*
     (14209 "When this variable is checked, style information is retained when text is pasted.")
     ; *load-verbose*
     (14210 "When this variable is checked, MCL prints a message to *standard-output* before it loads a file.")
     ; *verbose-eval-selection*
     (14211 "When this variable is checked, MCL prints the results of intermediate evaluations when a selection is evaluated. Otherwise, only the values of the final form are displayed.")
     ; *warn-if-redefine*
     (14212 "When this variable is checked, MCL issues a warning whenever any function, macro, or variable is redefined in a different file or buffer. Otherwise, warnings are issued only when built-in functions are redefined (unless *warn-if-redefine-kernel* is NIL).")
     ; *break-on-warnings*
     (14213 "When this variable is checked, MCL enters a break loop when a warning is signaled.")
     ; *backtrace-on-break*
     (14214 "When this variable is checked, MCL displays the Stack Backtrace dialog box whenever it enters a break loop. Otherwise, you must select Backtrace from the Tools menu to see the Stack Backtrace.")
     ; *break-on-errors*
     (14215 "When this variable is checked, MCL enters a break loop when an error is signaled. Otherwise, errors cause the computation to be aborted.")
     ; *autoload-lisp-package*
     (14216 "When this variable is checked, the :lisp package is automatically loaded when it is required. The :lisp package provides partial, but not complete backward compatibility with earlier versions of Common Lisp.")
     ; *fasl-save-local-symbols*
     (14217 "When this variable is checked, local symbols are saved in compiled files and are available when the files are loaded.  See *fasl-save-symbols*.")
     ; *fred-save-window-positions*
     (14218 "When this variable is checked, the window size, window position, scroll position, and current selection of Fred windows are saved when files are saved and restored when they are later reopened.")
     ; *autoclose-inactive-listeners*
     (14221 "When this variable is checked, listeners created for break loops will close automatically when the break loops are exited.")
     ; *always-eval-user-defvars*
     (14222 "When this variable is checked, DEFVAR forms executed from a listener or Fred buffer are treated as DEFPARAMETER forms (i.e., the value is used even if the variable was already bound).")
     |#
     ))

#|
(make-ccl-balloon-help)
|#
