;;-*- Mode: Lisp; Package: CCL -*-

;;	Change History (most recent first):
;;  3 1/17/95  akh  part-color method for popup so will be white if back-color
;;  (do not edit before this line!!)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  pop-up-menu.lisp
;;
;;
;;  Copyright 1989-1994 Apple Computer, Inc.
;;  Copyright 1995 Digitool, Inc.
;;
;;  this file implements pop-up menus, according to the Apple standard.
;;  it also shows how multiple-inheritance can be handy!
;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Mod History
;;start of retained text
;; menu-select inverts pull-down-menu before action rather than view-click-event-handler doing it after
;; (action can cause redraw which makes parity wrong - and it looks better)
;; highlight-title is an initarg
;; crop titles that don't fit, do menu-item-action with event-processing enabled 
;; set-pop-up-menu-default-item does invalidate-view stead of most callers, set-view-size and position
;; don't if same making resizing dialogs less bouncy. 
;; 10/20/93 alice scribble on title of disabled menu as well as the menu
;; 10/16/93 alice menu-disable and enable for typein-menu do menu and editable-text
;; 10/15/93 alice view-draw-contents more tasteful when disabled and color or gray scale.
;;		and menu-select of typein-menu-menu puts back the top edge correctly when disabled.
;; 09/19/93 alice Add-menu-items leaves the default-item number unchanged and fixes the
;;	     check marks or sets it to 1 if any items and default-item previously 0.
;;---------------
;;start of added text
;;
;; 06/07/94 bill  (method remove-menu-items :around (pop-up-menu))
;; -------------  3.0d13
;; 08/18/93 bill  with-clip-rect -> with-clip-rect-intersect
;; -------------- 3.0d12
;; 04/10/93 alice view-click-event-handler for pop-up-menu make sure he's still there
;;		   after menu-select
;; 02/15/93 alice title (the text to left) does not hilite and is not mouse sensitive
;;		default size of typein accounts for title
;; 02/03/93 alice pop-up is no longer a dialog-item - use menu-enable, menu-title etc.
;; 01/18/93 alice add pull-down-menu and typein-menu
;; 11/11/92 bill  Straz'es patch to add specify background color with :menu-body.
;; 11/05/92 alice view-default-size dtrt if no items and :sequence
;; 11/04/92 alice set-view-position does invalidate view - for ift, export a few
;;--------------  2.0
;; 03/23/92 bill  set-view-size needed to force an erase.
;;                menu-select now works correctly for hierarchical pop-up menus
;;                (which are not Human Interface Guidelines compliant).
;; -------------  2.0f3
;; 10/18/91 bill  optimize view-draw-contents a little.
;;                Adjust position of pop up menu
;; 10/15/91 bill  window-font -> view-font
;;                Add the little System 7 triangle.
;;--------------  2.0b3
;; 06/21/91 bill  wkf's mod: Add foreground color for titles of pop up menus.
;;--------------  2.0b2
 
;;;;;;;;;;;;;;;;;;
;;
;;  packages, proclamations, and requires
;;

(in-package :ccl)

(eval-when (:compile-toplevel :execute)
  (require 'toolequ))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (export '(pop-up-menu selected-item pop-up-menu-item-display
            pop-up-menu-default-item
            pop-up-menu-auto-update-default
            pull-down-menu
            typein-menu) :ccl))


; Hah - no more dialog item!

(defclass pop-up-menu (menu simple-view)
  ((width-correction :allocation :class :initform 0
                     :accessor pop-up-menu-width-correction)
   (menu-rect :initform nil :accessor pop-up-menu-rect)
   (title-rect :initform nil :accessor pop-up-menu-title-rect)
   (default-item :initarg :default-item :initform 1
                 :accessor pop-up-menu-default-item)
   (auto-update-default :initarg :auto-update-default :initform t
                        :accessor pop-up-menu-auto-update-default)
   (item-display :initarg :item-display :initform :selection
                 :accessor pop-up-menu-item-display)
   (cached-title :initform nil :accessor pop-up-menu-cached-title))
  (:default-initargs
    :view-position nil 
    :view-size nil
    :menu-title ""
    :view-font '("Chicago" 12 :plain)))

(defmethod set-pop-up-menu-default-item ((menu pop-up-menu) num)
  (let* ((old (pop-up-menu-default-item menu))
         (items (menu-items menu)))    
    (when (neq old num)
      (when (neq old 0)
        (set-pop-up-item-check-mark (nth (1- old) items) nil))
      (when (and num (neq num 0))
        (set-pop-up-item-check-mark (nth (1- num) items) t))
      (setf (pop-up-menu-default-item menu) num)
      (when (eq (pop-up-menu-item-display menu) :selection)
        (invalidate-view menu)))))


(defmethod install-view-in-window ((menu pop-up-menu) dialog &aux ok)  
  (declare (ignore dialog))  
  (menu-install menu)
  (without-interrupts ; this is what dialog-item buys us
   (unwind-protect
     (let ((container (view-container menu)))
       (set-default-size-and-position menu container)
       (setq ok t))
     (unless ok
       (set-view-container menu nil))))
  (size-rectangles menu)
  (invalidate-view menu))


;;;;;;;;;;;;;;;;;;;;
;;
;;  definitions for pop-up menus
;;


(defmethod initialize-instance :after ((menu pop-up-menu)  &rest initargs &key highlight-title)
  (declare (ignore initargs))
  (when highlight-title (view-put menu :highlight-title t))
  (let ((default (pop-up-menu-default-item menu)))    
    (when (and default (neq default 0))
      (setq default (nth (1- default) (menu-items menu)))
      (when default (set-pop-up-item-check-mark default t)))))

(defmethod add-menu-items ((menu pop-up-menu) &rest items)
  (declare (ignore items))
  (call-next-method)
  (when (pop-up-menu-auto-update-default menu)
    (let* ((default (pop-up-menu-default-item menu))
           (items (menu-items menu)))
      (when items
        (dolist (item items)(set-menu-item-check-mark item nil))
        (cond ((and default (neq default 0))
               (set-pop-up-menu-default-item menu 0)
               (set-pop-up-menu-default-item menu default))
              (t (set-pop-up-menu-default-item menu 1)))))))


(defmethod set-view-size  ((menu pop-up-menu) h &optional v)  
  (declare (ignore h v))
  (let ((old-size (view-size menu)))
    (call-next-method)
    (when (not (eql old-size (view-size menu)))
      (size-rectangles menu)
      (setf (pop-up-menu-cached-title menu) nil)
      (invalidate-view menu t))))

(defmethod set-view-position  ((menu pop-up-menu) h &optional v)
  ;(declare (ignore h v))
  (when (not (eql (view-position menu) (make-point h v)))
    (call-next-method)
    (size-rectangles menu)
    (invalidate-view menu t)))

(defmethod size-rectangles ((menu pop-up-menu))
  "does a lot of tweaking to get the thing to draw right"
  (let* ((my-pos (view-position menu))
         (my-size (view-size menu)))
    (when (and my-pos my-size)
      (setq my-size (add-points my-size #@(-1 -1)))
      (let* ((text (menu-title menu))
             (title-offset (make-point (if (eql 0 (length text))
                                         0
                                         (+ 8 (string-width
                                               text
                                               (or (view-font menu)
                                                   (view-font (view-window menu))))))
                                       0))
             (menu-rect (or (pop-up-menu-rect menu)
                            (setf (pop-up-menu-rect menu) (make-record :rect))))
             (title-rect (or (pop-up-menu-title-rect menu)
                             (setf (pop-up-menu-title-rect menu)
                                   (make-record :rect)))))
        (rset menu-rect :rect.topleft (add-points my-pos title-offset))
        (rset menu-rect :rect.bottomright (add-points my-pos my-size))
        (rset title-rect :rect.topleft (add-points my-pos #@(0 1)))
        (rset title-rect :rect.bottomright (make-point (+ (point-h my-pos)
                                                          title-offset)
                                                       (+ (point-v my-pos)
                                                          (point-v my-size)
                                                          -2)))))))

(defmethod remove-view-from-window ((menu pop-up-menu))
  (menu-deinstall menu)
  (call-next-method)
  (without-interrupts
   (dispose-record (pop-up-menu-rect menu) :rect)
   (setf (pop-up-menu-rect menu) nil)
   (dispose-record (pop-up-menu-title-rect menu) :rect)
   (setf (pop-up-menu-title-rect menu) nil)))

(defmethod view-draw-contents ((menu pop-up-menu) &aux (items (menu-items menu)))
  (let* (;(pos (view-position menu))
         (text (menu-title menu)) ;(dialog-item-text menu))
         (ti-rect (pop-up-menu-title-rect menu))
         (no-title (equal text ""))
         (item-display (pop-up-menu-item-display menu))
         (enabled (menu-enabled-p menu))
         (colorp (color-or-gray-p menu))
         (pull-down-p (pull-down-menu-p menu)))
    (with-focused-dialog-item (menu)  ; take font from item, draw in containers coords - this is the other thing that dialog item gives us
    (multiple-value-bind (a d w leading)(view-font-codes-info menu)
      (declare (ignore a))
      (rlet ((a-rect :rect))
        (copy-record (pop-up-menu-rect menu) :rect a-rect)
        (let ((mi-title (cond ((eq item-display :selection)
                                      (let ((selection (pop-up-menu-default-item menu)))
                                        (cond ((null items) "<No Items>")
                                              ((zerop selection) "<No selection>")
                                              (t (menu-item-title
                                                  (nth (- selection 1) items))))))
                                     ((stringp item-display)
                                      item-display)
                                     (t 
                                      (format nil "~a" item-display)))))
          (with-fore-color (if (and (not enabled) colorp)
                             *gray-color*
                             (part-color menu :menu-title)) ; 21-Jun-91 -wkf
            (with-back-color (part-color menu :menu-body) ; 10-Nov-92 -straz
              (unless no-title
                (#_EraseRect :ptr ti-rect)
                (#_MoveTo :word (+ (rref ti-rect rect.left) 3) ; (+ (point-h pos) 3)
                 :word (- (rref ti-rect rect.bottom) (+ d leading)))
                (with-pstrs ((di-title text))
                  (#_DrawString :ptr di-title)))
              ;  (#_OffsetRect :ptr a-rect :long #@(0 -1))
              (#_FillRect :ptr a-rect :ptr *white-pattern*)
              (cond ((not pull-down-p)
                     (#_FrameRect :ptr a-rect)
                     (#_MoveTo :word (+ (rref a-rect rect.left) 3)
                      :word (rref a-rect rect.bottom))
                     (#_LineTo :word (rref a-rect rect.right)
                      :word (rref a-rect rect.bottom))
                     (#_LineTo :word (rref a-rect rect.right)
                      :word (+ (rref a-rect rect.top) 2))))
              (#_InsetRect :ptr a-rect :long #@(1 1))
              (let* ((left (+ (rref a-rect rect.left)(if pull-down-p 6 (max 6 w))))
                     (right (rref a-rect rect.right))
                     (bottom (rref a-rect rect.bottom)))                    
                (#_MoveTo :word left :word  (- bottom (+ leading 1 d)))
                (with-clip-rect-intersect a-rect
                  (draw-string-crop mi-title (- right left (if pull-down-p 0 12)))
                  (#_MoveTo :word (- right (+ 4 11))
                   :word (- (ash (+ bottom (rref a-rect :rect.top)) -1)
                            2)))
                ; Draw the little triangle.
                (unless pull-down-p
                  (#_line :long #@(10 0))
                  (#_line :long #@(-5 5))
                  (#_line :long #@(-4 -4))
                  (#_line :long #@(7 0))
                  (#_line :long #@(-3 3))
                  (#_line :long #@(-2 -2))
                  (#_line :long #@(3 0))
                  (#_line :long #@(-1 1)))))))))
      (unless (or enabled colorp)
        (rlet ((ps :penstate))
          (with-item-rect (rect menu)
            (#_InsetRect :ptr rect :long #@(0 -1))
            (#_GetPenState :ptr ps)
            (#_PenPat :ptr *gray-pattern*)
            (#_PenMode :word 11)
            (#_PaintRect :ptr rect)
            (unless no-title (#_PaintRect ti-rect)) ; ??
            (#_SetPenState :ptr ps)))))))

(defmethod part-color ((menu menu) key)
  (or (getf (slot-value menu 'color-list) key nil)
      (case key (:menu-body *white-color*))))

(defmethod menu-disable ((menu pop-up-menu))
  (invalidate-view menu)
  (call-next-method))

(defmethod menu-enable ((menu pop-up-menu))
  (invalidate-view menu)
  (call-next-method))

(defmethod view-default-size ((menu pop-up-menu))
  (multiple-value-bind (ff ms)(view-font-codes menu)
    (let* ((item-display (slot-value menu 'item-display))
           (max-menu-width (max 10 (if (stringp item-display)
                                     (font-codes-string-width item-display ff ms)
                                     (if (and (eq item-display :selection)
                                              (not (menu-items menu)))
                                       (font-codes-string-width  "<no items>" ff ms)
                                       0))))
           (title (menu-title menu))
           (title-width (if (eq 0 (length title)) 0 (font-codes-string-width title ff ms))))
      ;(when ff (setq ff (logand ff #xffff0000))) ; nuke boldness if any
      (setq max-menu-width
            (+ (if (eq 0 title-width) 9 18)
               ; we used to dolist always
               (if (eq item-display :selection)
                 (let ((item-max (max-menu-width menu)))
                   (if (> item-max max-menu-width)
                     item-max
                     max-menu-width))
                 max-menu-width)))
      (multiple-value-bind (a d w l)(font-codes-info ff ms)
        (make-point (+  title-width  max-menu-width (if (pull-down-menu-p menu) 5 (+ 12 w))) ; was 15
                    (+ a d l 4))))))
#|
(defmethod point-in-click-region-p ((menu pop-up-menu) where)
  ; prevent changing selection when mouse doesn't move.
  ; should clicking in the title really select the menu? - naah
  (when (view-contains-point-p menu where)
    (let ((lh (view-font-line-height menu))
          (vh (point-v (view-size menu))))
      (declare (fixnum lh vh))
      (multiple-value-bind (floor rem) (floor (- vh lh 2) 2)
        ; this assumes that the v-offset is fancier than it is
        (let* ((rect (pop-up-menu-rect menu))
               (v-where (point-v where))
               (h-where (point-h where))
               (left (rref rect :rect.left))
               (right (rref rect :rect.right)))
          (and (< left h-where right)
               (let ((top (rref rect :rect.top)))
                 (and (> v-where (+ floor rem top))
                      (< v-where (+ top lh floor))))))))))
|#

(defmethod point-in-click-region-p ((menu pop-up-menu) where)
  ; prevent changing selection when mouse doesn't move.
  (when (view-contains-point-p menu where)
    (let ((lh (view-font-line-height menu))
          (vh (point-v (view-size menu)))
          (rect (pop-up-menu-rect menu)))
      (declare (fixnum lh vh))
      (and (< (rref rect :rect.left) (point-h where)
              (- (rref rect :rect.right) (menu-display-h-offset menu)))
           (or (< vh (+ 2 lh))          
               (let* ((v-where (point-v where))
                      (v-pos (point-v (view-position menu)))
                      (offset (menu-display-v-offset menu)))
                 (and (> v-where (+ offset v-pos))
                      (< v-where
                         (+ v-pos lh offset)))))))))
  

(defmethod view-click-event-handler ((menu pop-up-menu) where)
  (declare (ignore where))
  (let ((no-text (eq 0  (length (menu-title menu)))))    
    (unless no-text
      ; want to hilite not invert
      (title-hilite menu))
    (menu-select menu 0)        
    (unless no-text
      (title-hilite menu t))))


(defmethod title-hilite ((menu pop-up-menu) &optional un)
  (when (view-get menu :highlight-title)
    (let ((rect (pop-up-menu-title-rect menu)))
      (when rect
        (if un (invalidate-view menu)
            (with-macptrs ((hilitemode (%int-to-ptr #$HiliteMode)))    
              (%put-byte hilitemode (%ilogand2 #x7f (%get-byte hilitemode)))
              (#_InvertRect rect)))))))
#|
(defmethod title-hilite ((menu pop-up-menu) &optional un)
  (when *hilite-title*
    (let ((tc (part-color menu :menu-title))
          (bc (part-color menu :menu-body))
          (text (menu-title menu))
          (ti-rect (pop-up-menu-title-rect menu)))    
      (if (and tc bc)
        (let* ((a-rect (pop-up-menu-rect menu))
               (pos (with-focused-view (view-container menu)
                      (%local-to-global 
                       (wptr menu)
                       (rref a-rect :rect.topleft)))))        
          (with-fore-color (if un tc bc)
            (with-back-color (if un bc tc)
              (#_EraseRect :ptr ti-rect)
              (#_MoveTo :word (+ (point-h pos) 3)
               :word (- (rref a-rect rect.bottom) 8))
              (with-pstrs ((di-title text))
                (#_DrawString :ptr di-title)))))
        (#_InvertRect :ptr ti-rect)))))
|#

;Update the menu's items then displays the pop-menu.  Default-item is the
;item which will come up selected  when the menu is displayed.
(defmethod menu-select ((menu pop-up-menu) num
                        &aux selection
                        selected-menu
                        selected-menu-item
                        (a-rect (pop-up-menu-rect menu))
                        (pos (with-focused-view (view-container menu)
                               (%local-to-global 
                                (wptr menu)
                                (rref a-rect :rect.topleft)))))
  (declare (ignore num))
  (menu-update menu)
  (multiple-value-bind (ff ms)(view-font-codes menu)
  (let* ((handle (menu-handle menu))
         (cached-title (pop-up-menu-cached-title menu))
         (items (menu-items menu))
         (orig (if items (menu-item-title (car items))))
         (font (lsh ff -16))
         (font-size (logand ms #xffff))
         (sysfont (%get-word (%int-to-ptr $sysfontfam)))
         (sys-size (%get-word (%int-to-ptr $sysfontsize)))
         (same-p  (and (= font sysfont)
                       (or (eq sys-size font-size)
                           (and (zerop sys-size)(eq font-size 12))))))         
    (when (and (not cached-title) items)
      (setq cached-title (adjusted-menu-item-title menu)))
    (unwind-protect
      (let ()
        (when cached-title (set-menu-item-title (car items) cached-title))
        (unless same-p
          (setf (%get-word (%int-to-ptr $sysfontfam)) font)
          (setf (%get-word (%int-to-ptr $sysfontsize)) font-size)
          (setf (%get-long (%int-to-ptr $LastSPExtra)) -1))
        (setq selection (#_PopUpMenuSelect
                     :ptr handle
                     :word (+ (point-v pos) (menu-display-v-offset menu))
                     :word (+ (point-h pos) (menu-display-h-offset menu))
                     :word (or (pop-up-menu-default-item menu) 0)
                     :long)
              ;we get the selected menu in case you want to break the rules
              ;and use heirarchical menus in a pop-up menu
              selected-menu (menu-object (ash (logand #xFFFF0000 selection) -16))
              selected-menu-item (logand #x0000FFFF selection)))
      (unless  same-p
          (setf (%get-word (%int-to-ptr $sysfontfam)) sysfont)        
          (setf (%get-word (%int-to-ptr $sysfontsize)) sys-size)
          (setf (%get-long (%int-to-ptr $LastSPExtra)) -1))          
      (when cached-title
        (set-menu-item-title (car items) orig)
        (setf (pop-up-menu-cached-title menu) cached-title)))
    (when (typep menu 'pull-down-menu) ; gag-puke
      (#_InvertRect :ptr (pop-up-menu-rect menu)))
    (unless (eq selected-menu-item 0)
      (let* ((items (menu-items selected-menu))
             (update (pop-up-menu-auto-update-default menu)))
        (when update
          (set-pop-up-menu-default-item menu
                (if (eq selected-menu menu)
                  selected-menu-item
                  (let ((1st-level-submenu selected-menu))
                    (loop
                      (let ((owner (menu-owner 1st-level-submenu)))
                        (if (eq owner menu)
                          (return (1+ (position 1st-level-submenu (menu-items menu)))))
                        (if (null owner)
                          (return (pop-up-menu-default-item menu)))
                        (setq 1st-level-submenu owner))))))
          #|(when (eq (pop-up-menu-item-display menu) :selection)
            (invalidate-view menu))|# )
        (with-event-processing-enabled 
          (when update (event-dispatch))  ; let it be drawn          
          (menu-item-action
           (nth (- selected-menu-item 1) items))))))))



(defmethod menu-display-v-offset ((menu pop-up-menu))
  1)


(defmethod menu-display-h-offset ((menu pop-up-menu))
  1)
  

(defmethod menu-install ((menu pop-up-menu))
  "Creates the actual Macintosh menu with all of the menu's current items."
  (let* ((menu-items (menu-items menu)))
    (apply #'remove-menu-items menu menu-items)
    (init-menu-id menu)
    (with-pstrs ((menu-title (menu-title menu)))
      (let ((menu-handle (#_NewMenu :word (slot-value menu 'menu-id)
                                   :ptr menu-title
                                   :ptr)))
        (#_InsertMenu :ptr menu-handle
                     :word -1)
        (setf (slot-value menu 'menu-handle) menu-handle)))
    (let* ((colors (part-color-list menu)))
      (loop
        (unless colors (return))
        (set-part-color menu (pop colors) (pop colors))))
    (apply #'add-menu-items menu menu-items)))

(defmethod menu-deinstall ((menu pop-up-menu))
  (let* ((*menubar-frozen* t))
    (call-next-method)))

(defmethod selected-item ((menu pop-up-menu))
  (nth (- (pop-up-menu-default-item menu) 1) (menu-items menu)))

; This prevents the default item from becoming larger than the
; number of items in the menu. It also keeps the default-item
; constant even if its position changes.
(defmethod remove-menu-items :around ((menu pop-up-menu) &rest items)
  (declare (ignore items))
  (let* ((default-item-number (pop-up-menu-default-item menu))
         (default-item (unless (eql default-item-number 0)
                         (nth (1- default-item-number) (menu-items menu)))))
    (unwind-protect
      (call-next-method)
      (when default-item
        (let ((index (position default-item (menu-items menu))))
          (setf (pop-up-menu-default-item menu)
                (if index (1+ index) 1)))))))

;;;;;;;;;;;;;;;;;
;; pull-down-menu omits the triangle and outline. No default

(defclass pull-down-menu (pop-up-menu)
  ()
  (:default-initargs
    :auto-update-default nil
    :default-item 0))

(defmethod menu-display-v-offset ((menu pull-down-menu))
  (1- (point-v (view-size menu))))

(defun pull-down-menu-p (menu)
  (typep menu 'pull-down-menu))

(defmethod view-click-event-handler ((menu pull-down-menu) where)
  (declare (ignore where))
  (#_InvertRect :ptr (pop-up-menu-rect menu))
  (menu-select menu 0))

(defmethod point-in-click-region-p ((menu pull-down-menu) where)
  ; in this case it can be anyplace
  (view-contains-point-p menu where))
  

;;;;;;;;;;;;;;;;;
;;
;; typein-menu - its items should be of type typein-menu-item
;;


(defclass typein-menu (view)
  ((menu :initform nil :accessor typein-menu-menu)
   (menu-position :initarg :menu-position :initform :right :reader typein-menu-menu-position) 
   (editable-text :initform nil
                  :accessor typein-editable-text))
  (:default-initargs
    :menu-class 'typein-menu-menu
    :view-subviews nil
    :editable-text-class 'editable-text-dialog-item))

(defmethod menu-disable ((menu typein-menu))
  (let ((mm (typein-menu-menu menu)))
    (when  (menu-enabled-p mm)
      (menu-disable mm)
      (dialog-item-disable (typein-editable-text menu))
      (invalidate-view menu))))  ; get the title if any

(defmethod menu-enable ((menu typein-menu))
  (let ((mm (typein-menu-menu menu)))
    (unless (menu-enabled-p mm)
      (menu-enable mm)
      (dialog-item-enable (typein-editable-text menu))
      (invalidate-view menu))))

(defclass typein-menu-menu (pop-up-menu)()
  (:default-initargs
    ;:auto-update-default nil
    :item-display ""
    :default-item 0))

(defmethod initialize-instance :after ((view typein-menu-menu) &rest initargs &key &allow-other-keys)
  (declare (ignore initargs)))


(defmethod menu-display-v-offset ((menu typein-menu-menu))
  (case (typein-menu-menu-position (view-container menu))
    (:left 2) ; line up with title & text, doesn't cover the control (could erase it oneself)
    (t 2)))   ; maybe we will

(defmethod add-menu-items ((menu typein-menu-menu) &rest items)
  (let* ((second (second items))
         (typein-p (and second (string= (menu-item-title second) "-"))))
    (if (not typein-p)
      (apply #'call-next-method
             menu
             (make-instance 'typein-menu-item :menu-item-title "None")
             (make-instance 'menu-item :menu-item-title "-")
             items)
      (call-next-method))))

; do we really want these dudes to have titles?
(defmethod size-rectangles ((menu typein-menu-menu))
  (let ((title (menu-title menu)))
    (if (eq (length title) 0)
      (call-next-method)
      (let* ((my-pos (view-position menu))
             (my-size (view-size menu)))
        (when (and my-pos my-size)
          (multiple-value-bind (ff ms)(view-font-codes menu)
            (setq my-size (add-points my-size #@(-1 -1)))
            (let* ((menu-rect (or (pop-up-menu-rect menu)
                                  (setf (pop-up-menu-rect menu) (make-record :rect))))
                   (title-rect (or (pop-up-menu-title-rect menu)
                                   (setf (pop-up-menu-title-rect menu)
                                         (make-record :rect))))
                   (title-width (+ 8 (font-codes-string-width title ff ms))))
              (rset menu-rect :rect.topleft my-pos)
              (rset menu-rect :rect.bottomright (add-points my-pos my-size))
              (rset title-rect :rect.topleft (make-point 0 (+ (point-v my-pos) 2)))
              (rset title-rect :rect.bottomright (make-point title-width
                                                             (+ (point-v my-size)
                                                                (point-v my-pos)
                                                                -1))))))))))

(defmethod initialize-instance ((view typein-menu) &rest initargs
                                &key view-size
                                view-position
                                menu-class
                                menu-position
                                menu-items
                                menu-title                                
                                view-font
                                item-display
                                (draw-outline -2)
                                (dialog-item-text "")
                                editable-text-class)
  (declare (dynamic-extent initargs)(ignore menu-position menu-items menu-title item-display))
  (apply #'call-next-method view 
         :view-size view-size    ; make default size&pos work right
         :view-position view-position
         initargs)
  (let ((menu (apply #'make-instance menu-class               
                     :view-container view
                     initargs)))
    (setf (typein-menu-menu view) menu)    
    (let* ((edit (make-instance editable-text-class
                  :view-container view                  
                  :draw-outline draw-outline
                  (if view-font :view-font :ignore) view-font
                  ;:margin  (max (nth-value 2 (font-codes-info ff ms)) 6)
                  :dialog-item-text dialog-item-text
                  :allow-returns nil :allow-tabs nil)))
      (setf (typein-editable-text view) edit))
    (when view-size (view-size-parts view))))

; if fonts differ the height should be max of font heights
(defmethod view-default-size ((view typein-menu))
  (multiple-value-bind (ff ms)(view-font-codes view)
    (let* ((text (typein-editable-text view))
           (size (or (view-size text) (view-default-size text)))
           (h (max 100 (point-h size)))
           (title (menu-title (typein-menu-menu view)))           
           (title-width (if (eq 0 (length title)) 0 (font-codes-string-width title ff ms))))
      (make-point (+ h title-width 28)(+ 4 (point-v size))))))


(defmethod view-size-parts ((view typein-menu))  
  (let* ((size (view-size view))
         (size-v (point-v size))
         (size-h (point-h size))
         (menu (typein-menu-menu view))
         (text (typein-editable-text view))
         (title (menu-item-title menu))
         (title-width 0))
    (multiple-value-bind (ff ms)(view-font-codes view)
      (when (not (eql 0 (length title)))        
        (setq title-width (+ 8 (font-codes-string-width title ff ms))))
      (let* ()      
        (set-view-size menu (make-point (+ 22 0)  size-v))
        (set-view-size text (make-point (- size-h (+ 28 title-width)) (- size-v 4)))
        (case (typein-menu-menu-position view)
          (:left (set-view-position menu (make-point title-width 0))
                 (set-view-position text (make-point (+ 26 title-width ) 2)))
          (t 
           (set-view-position menu (make-point (+  size-h -22) 0))
           (set-view-position text (make-point (+ 2 title-width)  2))))))))



(defmethod set-default-size-and-position ((view typein-menu) &optional container)
  (declare (ignore container))
  (call-next-method)
  (view-size-parts view))
    

(defmethod install-view-in-window ((menu typein-menu) dialog &aux ok)  
  (declare (ignore dialog))
  (without-interrupts
   (unwind-protect
     (let ((container (view-container menu)))
       (set-default-size-and-position menu container)
       (setq ok t))
     (unless ok
       (set-view-container menu nil))))
  (call-next-method))

; do we want to  deal with vertical changes as well? naah
(defmethod set-view-size ((view typein-menu) h &optional v)
  (declare (ignore h v))
  (let ((menu (typein-menu-menu view)))
    (setf (pop-up-menu-cached-title menu) nil)
    (call-next-method)
    (view-size-parts view)
    ;(invalidate-view view)
    )
  (view-size view))


(defmethod menu-display-h-offset ((menu typein-menu-menu))
  (case (typein-menu-menu-position (view-container menu))
    ((:right :left) 
     1)
    (t  ; this we probably won't ever use     
     (let* ((text-size (point-h (view-size (typein-editable-text 
                                            (view-container menu))))))       
       (- -5 text-size)))))

(defun max-menu-width (menu)
  (multiple-value-bind (ff ms)(view-font-codes menu)
    ;(setq ff (logand ff #Xffff0000))
    (let* ((max 0))
      (dolist (m (menu-items menu) max)
        (when (> (setq m (font-codes-string-width (menu-item-title m)
                                                  ff ms))
                 max)
          (setq max m))))))

#|
; is this worth the trouble/space????
(defmethod menu-select ((menu typein-menu-menu) num)
  (declare (ignore num))
  (let ((num (pop-up-menu-default-item menu)))
    (if (and num (> num 1))
      (call-next-method)
      (let ((pos (typein-menu-menu-position (view-container menu)))
            (view-pos (view-position menu)))
        (rlet ((rect :rect)
               (t-rect :rect))
          (rset rect :rect.topleft view-pos)
          (rset rect :rect.bottom (1+ (point-v view-pos)))
          (rset rect :rect.right (+ -1 
                                    (point-h view-pos)
                                    (point-h (view-size menu))))
          (#_eraserect rect)
          (when  (eq pos :left)
            (let* ((text (typein-editable-text (view-container menu)))
                   (tpos (view-position text)))
              (rset t-rect :rect.topleft (subtract-points tpos #@(2 2)))
              (rset t-rect :rect.bottom (- (point-v tpos) 1))
              (rset t-rect :rect.right (+ 2 (point-h tpos)(point-h (view-size text))))
              (#_eraserect t-rect)))
          (call-next-method)
          (#_framerect rect)
          (when  (eq pos :left)(#_framerect t-rect)))))))
|#

(defmethod menu-select ((menu typein-menu-menu) num)
  (declare (ignore num))
  (let ((num (pop-up-menu-default-item menu)))
    (if (and num (> num 1))
      (call-next-method)      
      (let* ((c (view-container menu))
             (pos (typein-menu-menu-position c))
             (view-pos (view-position menu))
             (w (- (point-h (view-size menu)) 2))
             (tl view-pos)
             (br (make-point (point-h (view-size c)) 1))
             t-pos t-w)
        (when (eq pos :left)
          (let* ((text (typein-editable-text c)))
            (setq t-pos (subtract-points (view-position text) #@(2 2)))
            (setq t-w (+  (point-h (view-size text)) 3))
            (setq tl #@(0 0))))
        (rlet ((rect :rect :topleft tl :bottomright br))
          ; erase top edge which is not obscured by the menu contents
          (#_eraserect rect)
          (call-next-method)
          ; restore top edge
          (if (menu-enabled-p menu)
            (progn                
              (#_moveto :long view-pos)
              (#_line :word w :word 0)
              (when (eq pos :left)
                (#_moveto :long t-pos)
                (#_line :word t-w :word 0)))
            (invalidate-corners c tl br)))))))
    

; vanilla menu-item action functions do not take an argument!

(defclass typein-menu-item (menu-item) ())

(defmethod menu-item-action ((item typein-menu-item))
  (let ((action (menu-item-action-function item)))    
    (let* ((menu (menu-item-owner item))
           (ti (typein-editable-text (view-container menu)))
           (new-text (if (and (eq 1 (pop-up-menu-default-item menu))
                              (string= "None" (menu-item-title item)))
                       "" 
                       (menu-item-title item))))
      (when (not (string= (dialog-item-text ti)
                          new-text)) ; prevents flashies
        (set-dialog-item-text ti new-text))
      (let ((items (menu-items menu)))
        (dolist (i items)
          (when (menu-item-check-mark i)
            (set-pop-up-item-check-mark i nil)
            (return))))
      (set-pop-up-item-check-mark item t))
    (if action (funcall action item))))

(defun set-pop-up-item-check-mark (item mark)
  (let ((menu (menu-item-owner item)))
    (when (and menu (eq mark t)
               (neq (lsh (view-font-codes menu) -16)
                    #.(lsh (font-codes '("chicago")) -16))) ; you really mean is it chicago
      (setq mark #\dot)) ; or #\altCheckMark
    (set-menu-item-check-mark item mark)))
  


(defmethod view-click-event-handler ((menu typein-menu-menu) where)
  (declare (ignore where))  
  (let* ((ti (typein-editable-text (view-container menu)))
         (items (menu-items menu))
         (text-p (not (= 0 (dialog-item-text-length ti)))))
    (when (not items)
      (add-menu-items menu)
      (setq items (menu-items menu)))
    (if text-p
      (let ((str (dialog-item-text ti)))
        (unless
          (let ((n 1))
            (dolist (item (menu-items menu) nil)
              (when (string-equal str (menu-item-title item))
                (set-pop-up-menu-default-item menu n)
                (return t))
              (setq n (1+ n))))
          (setf (pop-up-menu-cached-title menu) nil)
          (set-menu-item-title (car items) str)
          (set-pop-up-menu-default-item menu 1)))
      (progn
        (when (not (string= (menu-item-title (car items)) "None"))
          (setf (pop-up-menu-cached-title menu) nil)          
          (set-menu-item-title (car items) "None"))
        (set-pop-up-menu-default-item menu 1)))
    (with-focused-view (view-container menu)
      (call-next-method))        
    (when (dialog-item-enabled-p ti)
      (set-current-key-handler (view-window menu) ti))))


(defmethod adjusted-menu-item-title ((menu pop-up-menu))
  ; gag
  (let ((items (menu-items menu)))
    (when  items
      (multiple-value-bind (ff ms)(view-font-codes menu)
        ;(setq ff (logand ff #xffff0000))
        (let* ((w (nth-value 2 (font-codes-info ff ms)))
               (max (+ w 12 (max-menu-width menu)))
               (width (menu-body-width menu)))
          ;(print (list max width))
          (when  (< max width)
            (setq max (+ w 13 (font-codes-string-width (menu-item-title (car items)) ff ms)))
            (let* ((ss (font-codes-string-width " " ff ms))
                   (first (car items)))           
              (%str-cat (menu-item-title first)
                        (make-array (ceiling (- width max) ss)
                                    :element-type 'base-character
                                    :initial-element #\space)))))))))


(defmethod menu-body-width ((menu pop-up-menu))
  (let* ((ti-rect (pop-up-menu-title-rect menu))
         (ti-width (- (rref ti-rect rect.right)(rref ti-rect rect.left))))
    (- (point-h (view-size menu))  ti-width)))

(defmethod menu-body-width ((menu typein-menu-menu))
  (let* ((ti-rect (pop-up-menu-title-rect menu))
         (ti-width (- (rref ti-rect rect.right)(rref ti-rect rect.left))))
    (-  (point-h (view-size (view-container menu))) ti-width)))

(defmethod adjusted-menu-item-title ((menu pull-down-menu))
  (declare (ignore menu)))

(defmethod adjusted-menu-item-title ((menu typein-menu-menu))
  (case (typein-menu-menu-position (view-container menu))
    (:right nil)
    (t (call-next-method))))

(defmethod add-menu-items :after ((menu pop-up-menu) &rest items)
  ;(declare (ignore items))
  ; fix this mess or nuke it - looks cooler plain  
  (let ((style (inverse-style-arg (lsh (logand  (view-font-codes menu)  #xffff) -8))))
      (when (neq style :plain)
        (dolist (i items)
          (set-menu-item-style i style))))  
  (setf (pop-up-menu-cached-title menu) nil))

(defmethod remove-menu-items :after ((menu pop-up-menu) &rest ignore)
  (declare (ignore ignore))
  (setf (pop-up-menu-cached-title menu) nil))

(provide 'pop-up-menu)

#|

; a "menubar" in a dialog
(defclass my-menubar (inspector::bottom-line-mixin view)())


(defmethod view-default-size ((view my-menubar))
  (let ((container (view-container view)))        
    (when container       
      (make-point (point-h (view-size container))
                  (+ (view-font-line-height view) 4)))))

(defmethod install-view-in-window ((view my-menubar) w)
  (declare (ignore w))
  (multiple-value-bind (ff ms)(view-font-codes view)
    (let ((container (view-container view)))
      (when ff
        (do-subviews (sub view)
          (set-view-font-codes sub ff ms)))
      ; sets slots
      (set-default-size-and-position view container)
      (call-next-method)
      (invalidate-view view))))


(defun test ()
  (setq my-pop-up
        (make-instance 'pop-up-menu
          :menu-title "Wowieg"
          :highlight-title t
          :view-font '("geneva" 24 :bold)
          :menu-items
          (list
           (make-instance 'menu-item
             :menu-item-title "item oneg"
             :menu-item-action #'(lambda ()
                                   ))
           (make-instance 'menu-item
             :menu-item-title "item two"
             :menu-item-action #'(lambda ()
                                   (print 2)))
           (make-instance 'menu-item
             :menu-item-title "item three"
             :menu-item-action #'(lambda ()
                                   (print 3)))
           (make-instance 'menu-item
             :menu-item-title "item fourteen"
             :menu-item-action #'(lambda ()
                                   (print 14))))))
  
  (setq my-pop-up-2
        (make-instance 'pop-up-menu
          :item-display "Wowieg"
          
          :menu-items
          (list
           (make-instance 'menu-item
             :menu-item-title "item one"
             :menu-item-action #'(lambda ()
                                   ))
           (make-instance 'menu-item
             :menu-item-title "item two"
             :menu-item-action #'(lambda ()
                                   (print 2)))
           (make-instance 'menu-item
             :menu-item-title "item three"
             :menu-item-action #'(lambda ()
                                   (print 3)))
           (make-instance 'menu-item
             :menu-item-title "item fourteen"
             :menu-item-action #'(lambda ()
                                   (print 14))))))
  
  (setq my-pop-up-3
        (make-instance 'typein-menu
          :menu-title "abcdefg"
          :view-font '("geneva" 10 :bold)
          :menu-position :right  ; also try :left
          :menu-items
          (list (make-instance 'typein-menu-item                
                  :menu-item-title "elephant"
                  :menu-item-action #'(lambda (item)(declare (ignore item))(print 'babar)))
                (make-instance 'typein-menu-item                
                  :menu-item-title "peanut soup"
                  :menu-item-action #'(lambda (item)(declare (ignore item))(print 'babar))))))
  (setq my-pop-up-4
        (make-instance 'typein-menu
          :menu-title "ephabt"
          :view-font '("monaco" 9 )
          :menu-position :left  ; also try :left
          :menu-items
          (list (make-instance 'typein-menu-item                
                  :menu-item-title "elephant"
                  :menu-item-action #'(lambda (item)(declare (ignore item))(print 'babar)))
                (make-instance 'typein-menu-item                
                  :menu-item-title "peanut soup"
                  :menu-item-action #'(lambda (item)(declare (ignore item))(print 'babar))))))
  
  (let* ((bar (make-instance 'my-menubar
               :view-font '("geneva" 10)
               :view-subviews
                (list 
                 (make-instance 'pull-down-menu
                   :item-display "Commands"
                   :view-position #@(1 1)
                   :menu-items
                   (list (make-instance 'menu-item
                           :menu-item-title "Eat it")
                         (make-instance 'menu-item
                           :menu-item-title "Inspect it")
                         (make-instance 'menu-item
                           :menu-item-title "Send it to Mom")))))))
  
  
    (setq my-dial (make-instance 'dialog
                  :view-size #@(500 150)
                  :window-title "Pop-up Menu Test"
                  :view-subviews (list bar my-pop-up my-pop-up-2 my-pop-up-3 my-pop-up-4)))))


|#

#|
	Change History (most recent last):
	2	12/29/94	akh	merge with d13
|# ;(do not edit past this line!!)
