;; This is SLOW, but fairly thorough.
;; invoke via : (BINARY-STREAM-TEST <scratch-file-name>)

(defparameter *bin-strm-dflt-width* 16)
(defparameter *bin-strm-test-size* 1000)
(defparameter *num-accesses* 100)

(defun stream-test-out (file &key (width *bin-strm-dflt-width*) (size *bin-strm-test-size*))
  (let* ((element-type `(unsigned-byte ,width))
         (limit (ash 1 width)))
  (with-open-file (stream file :element-type element-type 
                          :direction :output
                          :if-exists :supersede)
    (do* ((i 0 (1+ i))
          (n 0 (mod (1+ n) limit)))
         ((= i size))
      (write-byte n stream)))))

(defun stream-test-in (file &key (width *bin-strm-dflt-width*) (size *bin-strm-test-size*))
  (let* ((element-type `(unsigned-byte ,width))
         (limit (ash 1 width)))
    (with-open-file (stream file :element-type element-type)
      (unless (= (file-length stream) size)
        (warn "Bogus file length for width ~d: should be ~d, was ~d."
               width size (file-length stream)))
        (do* ((i 0 (1+ i))
              (match 0 (mod (1+ match) limit)))
             ((= i size))
          (let* ((pos (file-position stream)))
            (unless (= i pos)
              (error "Bogus stream position on ~s: should be ~d, was ~d"
                     stream i pos)))
          (let* ((b (read-byte stream)))
            (unless (= match b)
              (return-from stream-test-in (list i match b)))))
        (dotimes (i *num-accesses*)
          (let* ((pos (random size))
                 (match (mod pos limit)))
            (file-position stream pos)
            (let* ((nowpos (file-position stream)))
              (unless (= nowpos pos)
                (error "Bogus stream position on ~s: should be ~d, was ~d"
                       stream pos nowpos)))
            (let* ((b (read-byte stream)))
              (unless (= match b)
                (return (list pos match b)))))))))


(defun binary-stream-test (filename)
  (do* ((element-size 1 (1+ element-size)))
       ((> element-size 32))
    (destructuring-bind (&optional i match b) 
                        (progn 
                          (stream-test-out filename :width element-size)
                          (stream-test-in filename :width element-size))
      (if i
        (format t "~&;~s ~s FAILED on byte ~d: expected ~d, got ~d"
                `(unsigned-byte ,element-size) *bin-strm-test-size* i match b)
        (format t "~&;~s ~s ok" `(unsigned-byte ,element-size) *bin-strm-test-size*)))))

(defun do-bs-test (&optional (filename "scratch-file"))
  (dolist (size '(1 2 3 4 5 10 20 33 99 100 101 255 256 257 1000 10000))
    (let ((*bin-strm-test-size* size))
      (binary-stream-test filename))))

; (do-bs-test)

; End of bfs-tests.lisp
