/*
File "plsym.c" - PPC PLSYM dcmd

Copyright 1990-1994 Apple Computer, Inc.
Copyright 1995 Digitool, Inc.


Modification History

 11/03/95 slh   PPC version
  5/19/95 slh   dcmdScroll initially
*/

#include <Types.h>
#include <LowMem.h>
#include <string.h>
#include <ndcd.h>
#include <ctype.h>
#include "lisp.h"
#include "printsupp.h"

#include "::constants.h"
#include "::macros.h"
#include "::lisp_globals.h"

void describe_symbol( ulong sym ) {
  PutCStr("Symbol ");
  sprint_hex( sym );
	PutCStr(": ");
  sprint_string( misc_ref(sym, 1), 0 );
	PutCStr("; Value: ");
	print_object( misc_ref(sym, 2) );
	PutCStr("Function: ");
	print_object( misc_ref(sym, 3) );
}

ulong skip_over_ivector( ulong start, ulong header ) {
  ulong	element_count = header_element_count(header),
				subtag = header_subtag(header),
				nbytes;
    
  if (subtag <= max_32_bit_ivector_subtag) {
    nbytes = element_count << 2;
  } else if (subtag <= max_8_bit_ivector_subtag) {
    nbytes = element_count;
  } else if (subtag <= max_16_bit_ivector_subtag) {
    nbytes = element_count << 1;
  } else if (subtag == subtag_double_float_vector) {
    nbytes = 4 + (element_count << 3);
  } else {
    nbytes = (element_count+7) >> 3;
  }
  return( start + (~7 & (nbytes + 4 + 7)) );
}

#define lglobal(g) (((ulong *) (lnil-fulltag_nil))[(g)])

ulong dump_heap( void ) {
  ulong		lnil, hstart, object, hend, value, nobjs = 0;
	OSErr		oserr = dcmdReadPowerPCRegister( R9Register, &hend );

	if ( oserr == noErr ) oserr = dcmdReadPowerPCRegister( R2Register, &lnil );
	if ( oserr != noErr ) {
		PutCStr( "Sorry, OSErr " );
		PutUDec( oserr );
		PutCLine( " occurred.");
		return;
	}

	object = hstart = lglobal(HEAP_START);
	
  while( object < hend ) {
		PutUHexZ(object, 8);
		PutCStr(": ");
		++nobjs;
    value = *((ulong *) object);
    if (fulltag_of(value) == fulltag_immheader) {
			print_object( object | ft_misc );
			object = skip_over_ivector( object, value );
		} else {
		  PutCStr(">>");
			print_object( value );
			object += 4;
			PutUHexZ(object, 8);
			PutCStr(": > ");
			print_object( *((ulong *) object)++ );
			object += 8;
    }
  }
  PutCStr("Total objects: ");
  PutUDec(nobjs);
  PutLine();
}

ulong find_symbol( unsigned char *pname ) {
  ulong		lnil, hstart, object, hend, nsyms = 0;
	OSErr		oserr = dcmdReadPowerPCRegister( R9Register, &hend );

	if ( oserr == noErr ) oserr = dcmdReadPowerPCRegister( R2Register, &lnil );
	if ( oserr != noErr ) {
		PutCStr( "Sorry, OSErr " );
		PutUDec( oserr );
		PutCLine( " occurred.");
		return 0;
	}

	object = hstart = lglobal(HEAP_START);

  while( object < hend ) {
    ulong	header = ((ulong *) object)[0];
		if ( header_subtag(header) == st_symbol ) {
			size_t len = *((char *)pname);
			ulong symname = ((ulong *) object)[1] - t_misc;
			ulong header  = ((ulong *) symname)[0];
			++nsyms;
#if 0
			if (nsyms == 100) {
				PutPStr( pname ); PutUDec(len); PutCStr(" "); sprint_hex(symname+4);
				PutUDec(header_element_count(header)); sprint_hex( 0xffff & memcmp( symname+4, pname+1, len )); PutLine(); return 0;
				return 0;
			}
#endif
			if ( (len == header_element_count( header ))
			     && !(0xFFFF & memcmp( pname+1, symname+4, len )) ) {
				return( object | t_misc );
			} else {
				object += 24;
			}
		} else if ( fulltag_of(header) == fulltag_immheader ) {
			object = skip_over_ivector( object, header );
		} else {
			object += 8;
		}
  }
	return nsyms;
}

pascal void CommandEntry (dcmdBlock* paramPtr) {
  short	delim=0;
  ulong	address = 0;
  Boolean ok;
  unsigned char pname[256];

  switch (paramPtr->request) {
    case dcmdInit:
      break;

    case dcmdHelp:
      dcmdDrawLine("\pPLSYM <symbol-or-pname>|*");
      dcmdDrawLine("\p   Displays information about <symbol-or-pname>");
#if 0
      dcmdDrawLine("\p   If parameter is enclosed in double quotes, finds symbol");
      dcmdDrawLine("\p     with matching pname & displays information about it.");
			dcmdDrawLine("\p   * indicates to print a heap dump. Currently broken.");
      dcmdDrawLine("\p   Otherwise, <symbol-or-pname> should be tagged as a symbol.");
#endif
      break;

    case dcmdDoIt:
      while (!delim) {
        switch( dcmdPeekAtNextChar() ) {
					case ' ':
						(void) dcmdGetNextChar();	/* eat it */
						break;
					case '*':
						address = dump_heap();
						delim = 1;
						break;
					case '\"':
						delim = dcmdGetNextParameter(pname);
						address = find_symbol(pname);
						delim = 1;
						break;
					default:
						delim = dcmdGetNextExpression( &address, &ok );
						if (!ok) {
							dcmdDrawLine("\pSyntax error");
							return;
						}
						delim = 1;
						break;
				}
			}
      if ( stagp( address, st_symbol ) ) {
      	describe_symbol( address );
      } else {
				PutUDec( address ); PutLine();
        //dcmdDrawLine("\pNot a symbol.");
      }
      break;
    }
}

// End of plsym.c
